(**********************************)
(*  TP 5 : Cryptographie par RSA  *)
(**********************************)
 

#open "num";;
#open"big_int";;

let modN = mod_num;;
#infix "modN";;
let quoN = quo_num;;
#infix "quoN";;
let sqrt_num x = num_of_big_int (sqrt_big_int (big_int_of_num x));;


let zero = num_of_int 0;;
let un = num_of_int 1;;
let deux = num_of_int 2;;
let quatre = deux +/ deux ;;

let random_num num = 
  let a = string_of_num (square_num num) in
  let n = (string_length a) / 2 in
  let b = sub_string a (n / 2) n in
  let rand = (num_of_int (random__int max_int) +/ num_of_string b) modN num in
    if rand </ zero then rand +/ num else rand;;


 
(*-------------------------*)
(*  Implémentation de RSA  *)
(*-------------------------*)


(* Outils arithmétiques *)

(* 1 *)
let rec pgcd a b =
  if a =/ zero then b else pgcd (b modN a) a;;

(* pgcd : num -> num -> num *)

(* 2 *)
let is_prime n =
  let i = ref deux in
  let pgcd_res = ref un in
    while !pgcd_res =/ un && !i <=/ sqrt_num n do
      pgcd_res := pgcd !i n;
      i := !i +/ un
    done;
    !pgcd_res =/ un;;

(* is_prime : num -> bool *)

(* 3 *)
let generate_prime inf range =
  let guess = ref (inf +/ (random_num range)) in
    while not is_prime !guess do
      guess := inf +/ (random_num range) done;
    !guess;;
(* generate_prime : num -> num -> num *)


(* 4 *)
let rec prime_with n =
  let r = deux +/ random_num (n -/ deux) in
    if pgcd r n =/ un then r else prime_with n;;

(* prime_with : num -> num *)

(* 5 *)
let rec extended_pgcd a b =
  if a modN b =/ zero then (zero,un)
  else let u,v = extended_pgcd b (a modN b) in
	 (v,u -/ v*/(a quoN b));; 
(* extended_pgcd : num -> num -> num * num *)


(* 6 *)

(* La fonction est ici donnée sous forme récursive terminale *)
let modular_exp m e n =
  let rec rec_term_mod_exp m e n r = 
    if e =/ zero then r
    else
      let m_2 = (square_num m) modN n in
	if e modN deux =/ zero
	then rec_term_mod_exp m_2 (e quoN deux) n r
	else rec_term_mod_exp m_2 (e quoN deux) n ((r */ m) modN n)
  in
    rec_term_mod_exp m e n un;;

(* modular_exp : num -> num -> num -> num *)



(* Implémentation du protocole *)


type key = Private of num * num | Public of num * num;;

(* 7 *)
let generate_keys inf range =
  let p = generate_prime inf range in
  let q = let guess = ref (generate_prime inf range) in
	    while p =/ !guess do guess := generate_prime inf range done;
    !guess in
  let n = p */ q in
  let phi_n = (p -/ un) */ (q -/ un) in
  let e = prime_with phi_n in
  let d = (phi_n +/ fst (extended_pgcd e phi_n )) modN phi_n in
    (* Il faut faire attention, les éléments d'un couple de Bézout peuvent être négatifs *)
    ( Private (d, n ), Public (e , n ));;
(* generate_keys : num -> num -> key * key *)


(* 8 *)

let encode priv message =
  match priv with
    | Private _ -> failwith "encode: expect public key for encoding"
    | Public (e , n) -> modular_exp message e n ;;

let decode pub message =
  match pub with
    | Public _ -> failwith "decode: expect private key for decoding"
    | Private (d, n) -> modular_exp message d n ;;

(* encode : key -> num -> num *)
(* decode : key -> num -> num *)

(* 9 *)
let encode_string pub s =
  match pub with
    | Private _ -> failwith " encode_string : expect public key"
    | Public (_, n) ->
	let base = num_of_int 256 in
	let res = ref zero in
	  for i = 0 to (string_length s)-1 do
	    res := base */ !res +/ num_of_int ( int_of_char s.[i])
	  done;
	  res := !res */ base;
	  (* !res , !res +1 et !res +2 ne peuvent être tous multiples de p ou q, donc on teste *)
	  if pgcd !res n =/ un then !res
	  else if pgcd (!res +/ un) n =/ un then !res +/ un else !res +/ deux;;


(* encode_string : key -> string -> num  *)

let decode_string n =
  let base = num_of_int 256 in
  let nb = ref (n quoN base) in (* on supprime le dernier caractère *)
  let taille = ref 1 in
    while !nb >=/ base do (* calcul de la taille *) 
      incr taille ;
      nb := !nb quoN base 
    done;
    nb := n quoN base;
    let s = make_string (!taille) ("?".[0]) in
      for i = !taille - 1 downto 0 do (* écriture de la chaîne *)
	s.[ i ] <- char_of_int (int_of_num (!nb modN base)); 
	nb := !nb quoN base
      done;
      s ;;

(* Et on teste le tout, sur un message pas trop grand *)
 
let test_encode message =
  let log_size = int_of_float (1.205 *. float_of_int ( string_length message)) in
  let size = num_of_string ("1e" ^ ( string_of_int (2 + log_size ))) in
  let priv , pub = generate_keys size (sqrt_num size ) in
  let nb = encode_string pub message in
  let enc = encode pub nb in
    priv,pub,enc;;

let test_decode message priv =
  let dec = decode priv message in 
    decode_string dec;;

let priv,pub,message = encode "coucou";;

decode message priv;;


(*---------------------------------*)
(*  Un meilleur test de primalité  *)
(*---------------------------------*)


(* Le test de Fermat *)
    
let rec fermat precision n =
  precision = 0 || (* pareil que « if precision = 0 then true else » *)
  let a = deux +/ random_num (n -/ quatre) in
    (modular_exp a (n -/ un) n =/ un && fermat ( precision - 1) n );;

(* fermat : int -> num -> bool *)


(* Le test de Rabin-Miller *)

(* 12 *)
let decompose n =
  if n =/ zero then zero , zero
  else
    let s = ref zero and t = ref n in
      while !t modN deux =/ zero do
	t := !t quoN deux;
	s := !s +/ un
      done;
      !s , !t ;;
(* decompose : num -> num * num *)

let rec search_square_root n a s =
  a =/ n -/ un || (* et non -1 car les résidus sont positifs *)
  (s >/ un && search_square_root n (( square_num a) modN n) (s -/ un ));;

let rabin_miller_step n s t = 
  let a = deux +/ random_num (n -/ quatre) in
  let a_t = modular_exp a t n in
    a_t =/ un || (* si a^t = 1, inutile de chercher des racines carrées de 1 *)
      search_square_root n a_t s ;;

let rec iter_rabin_miller precision n s t =
  precision = 0 ||
  ( rabin_miller_step n s t && iter_rabin_miller ( precision - 1) n s t );;
let rabin_miller precision n =
  if n modN deux =/ zero then n =/ deux
  else let s , t = decompose (n -/ un) in iter_rabin_miller precision n s t ;;


let new_generate_prime inf range =
  let guess = ref ( inf +/ (random_num range)) in
    while not rabin_miller 50 !guess do
      guess := inf +/ (random_num range)
    done;
    !guess ;;

let new_generate_keys inf range =
  let p = new_generate_prime inf range in
  let q = 
    let guess = ref (new_generate_prime inf range) in
      while p =/ !guess do guess := new_generate_prime inf range done;
      !guess in
  let n = p */ q in
  let phi_n = (p -/ un) */ (q -/ un) in
  let e = prime_with phi_n in
  let d = (phi_n +/ fst (extended_pgcd e phi_n )) modN phi_n in
    ( Private (d, n ), Public (e , n ));;


 (* 13 *)

(* La même chose, mais avec un message beaucoup plus long à coder, et qui passe en temps raisonnable *)

let test_encode2 message =
  let log_size = int_of_float (1.205 *. float_of_int ( string_length message)) in
  let size = num_of_string ("1e" ^ ( string_of_int (2 + log_size ))) in
  let priv, pub = new_generate_keys size (sqrt_num size ) in
  let nb = encode_string pub message in
  let enc = encode pub nb in
    priv,pub,enc;;

let test_decode2 message priv =
  let dec = decode priv message in 
    decode_string dec;;

let priv,pub,message = test_encode2 "Un message beaucoup plus long";;
test_decode message priv;;

(* Et puis pour le principe, on fait la course *)
#open "sys";;

let t1=time() in 
let priv,pub,message = test_encode "coucou" in
  time()-.t1;;

let t1=time() in 
let priv,pub,message = test_encode2 "un message beaucoup plus long, et pourtant ça va plus vite" in
  time()-.t1;;


(*-------------------*)
(*  Attaques de RSA  *)
(*-------------------*)


(* Algorithme naïf *)
(* 14 *)
let rec factorize_aux factors_list i n =
  if i >/ sqrt_num n then n :: factors_list
  else
    if n modN i =/ zero
  then factorize_aux ( i :: factors_list ) i (n quoN i)
  else factorize_aux factors_list ( i +/ un) n ;;

let factorize n = rev ( factorize_aux [ ] deux n );;



(* Algorithme rho de Pollard *)
(* 15 *)
let cycle_detect f u0 max_iter =
  let tortoise = ref (f u0) in
  let hase = ref (f (f u0)) in
  let nb_step = ref un in
    while !tortoise <>/ !hase && !nb_step < max_iter do
      tortoise:= f !tortoise ;
      hase:= f (f !hase)
    done;
    !nb_step =/ max_iter ;;

(* cycle_detect: (num -> num) -> num -> num -> bool *)

(* 16 *)
let rec rho_aux nb_iter acc f n =
  if nb_iter =/ zero || n =/ un then acc
  else
    let a = deux +/ (random_num (n -/ quatre )) in
    let tortoise = ref (f a) in
    let hase = ref (f (f a )) in
    let pgcd_res = ref un in
      while !pgcd_res =/ un && !tortoise <>/ !hase do
	pgcd_res := pgcd n (abs_num (!hase -/ !tortoise ));
	tortoise := f !tortoise ;
	hase := f ( f !hase)
      done;
      if !hase =/ !tortoise
      then rho_aux (nb_iter -/ un) acc f n
      else rho_aux nb_iter (!pgcd_res::acc) f (n quoN !pgcd_res);;

let rho f n =
  let cst = ceiling_num (( num_of_int 1178) // (num_of_int 1000)) in
    rev (rho_aux (cst */ sqrt_num (sqrt_num n)) [] f n );;

(* 17 *)
let casse message pub =
  match pub with
    |Private(_) -> failwith "pfffff, trop facile"
    |Public(e,n) ->
       let l=factorize n in
       let p=hd(l) and q=hd(tl(l)) in
       let phi_n = (p -/ un) */ (q -/ un) in
       let d = (phi_n +/ fst (extended_pgcd e phi_n )) modN phi_n in
	 test_decode message (Private(d,n));;

let _,pub,message = test_encode "bli";;
casse message pub;; 

(* Avec un caractère de plus, cela devient déjà des trop grosses clefs pour factoriser en temps raisonnable *)
let _,pub,message = test_encode2 "blih";;
let t1=time() in let _ = casse message pub in time() -. t1;; 


let list2 = function
  |a::[b] -> true
  |_ -> false;;

let casse2 message pub =
  match pub with
    |Private(_) -> failwith "pfffff, trop facile"
    |Public(e,n) ->
       let f c x = (x*/x +/ c) modN n in
       let c=ref (random_num n )in
       let l=ref (rho (f !c) n )in
	 while not(list2 !l) do
	   (* on peut aussi rajouter un compteur afin de ne pas trop faire d'essais *)
	   c:=random_num n;
	   l:= rho (f !c) n 
	 done;
	 let p=hd(!l) and q=hd(tl(!l)) in
	 let phi_n = (p -/ un) */ (q -/ un) in
	 let d = (phi_n +/ fst (extended_pgcd e phi_n )) modN phi_n in 
	   test_decode message (Private(d,n));;

(* La taille limite que je peux atteindre sur ma machine sans trop dépasser la minute... *)
let _,pub,message = test_encode2 "secret";;
casse2 message pub;;
